$Env:CONDA_EXE = "C:\ci\conda_1617142528204\_h_env\Scripts\conda.exe"
$Env:_CE_M = ""
$Env:_CE_CONDA = ""
$Env:_CONDA_ROOT = "C:\ci\conda_1617142528204\_h_env"
$Env:_CONDA_EXE = "C:\ci\conda_1617142528204\_h_env\Scripts\conda.exe"

Import-Module "$Env:_CONDA_ROOT\shell\condabin\Conda.psm1"
Add-CondaEnvironmentToPrompt